import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
import scipy as sp
import math

def pareto_frontier(x, y):
    points = []
    for i in range(len(x)):
        keep = True
        for j in range(len(x)):
            if (x[i] > x[j]) and (y[i] < y[j]): 
                keep = False
                break
        if keep:
            points.append([x[i],y[i]])

    return np.array(sorted(points))          

def diameter(graph):
    try:
        return nx.diameter(graph)
    except:
        graph = max(nx.connected_component_subgraphs(graph), key=len)
        return nx.diameter(graph)
    
def smooth(x,window_len=11,window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal 
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also: 

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")


    s=np.r_[x[window_len-1:0:-1],x,x[-2:-window_len-1:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=np.ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')

    y=np.convolve(w/w.sum(),s,mode='valid')
    return y


def degree_max_fn(graph): 
    return max([deg for _, deg in nx.degree(graph)])

def degree_mean_fn(graph): 
    return np.mean([deg for _, deg in nx.degree(graph)])